# encoding: utf-8
# frozen_string_literal: true

# multiRASM Assembler (pseudo-)instructions

module RASM_Instructions

extend RASM_Directives

def _make_p8x32a_instruction(opcode,condition,wr,wc,wz,immediate,dest,src)
    raise ArgumentError,"src out of range" unless (0..0x1ff).include?(src)
    raise ArgumentError,"dest out of range" unless (0..0x1ff).include?(dest)
    raise ArgumentError,"opcode out of range" unless (0..0x3f).include?(opcode)
    i = src + (dest<<9) + (opcode<<26) + (condition.p8x32a<<18)
    i |= 1<<22 if immediate
    i |= 1<<23 if wr
    i |= 1<<24 if wc
    i |= 1<<25 if wz
    return i
end

# Checks if a multi-long instruction can use native condition bits
def _native_condition_allowed?(condition,datalong)
    ((~datalong>>18)&15) | condition.p8x32a == 15
end

def _cogloc(val)
    if val.is_a? Symbol
        _cogloc($labels[val].loc!)
    elsif val.is_a? Integer
        con = $vm_constants_map[val]
        return con.to_s if con
        "cog[0x#{"%03x" % val}]"
    else 
        raise
    end
end

def _frame_c_min(address,condition,str)
    <<~CODE.gsub(/\n+\s*\n+/,"\n")
        #{_c_label(address)}: if(#{condition.c}) {
        #{str.gsub(/^/,"    ")}
        }
    CODE
end

def _frame_c(address,str,condition,wr,wc,wz,dest,src)
    str = <<~CODE
            #{src ?  "vmlong source = #{src};" : ""}
            #{dest ? "vmlong dest = #{dest};" : ""}
            bool newc,newnz;
                #{str}
            #{wr ? "#{dest} = dest;" : ""}
            #{wc ? "c = newc;" : ""}
            #{wz ? "nz = newnz;" : ""}
    CODE
    _frame_c_min(address,condition,str)
end

unless $rasm_options.switchcase
    def _c_dynjmp(address,target)
        <<~INDIRECT
            vmlong address = #{target};
            const void* ptr = search_phonebook(phonebook_index,phonebook_size,address);
            if(!ptr) return dynjmp_fault(#{address},address);
            goto *ptr;
        INDIRECT
    end
    def _c_pushstack(hereaddr,pushaddr)
        <<~PUSH
            vmstack[vmstackptr++] = &&#{_c_label(pushaddr)};
            if (vmstackptr >= VMSTACK_DEPTH)
                return stack_fault(#{hereaddr});
        PUSH
    end
    def _c_pullstack(hereaddr)
        <<~PULL
            if (--vmstackptr < 0)
                return stack_fault(#{hereaddr});
            goto *vmstack[vmstackptr];
        PULL
    end
else
    def _c_dynjmp(address,target)
        <<~INDIRECT
            entry = #{target};
            dyncallsite = #{address};
            goto reentry;
        INDIRECT
    end
    def _c_pushstack(hereaddr,pushaddr)
        <<~PUSH
            vmstack[vmstackptr++] = #{pushaddr};
            if (vmstackptr >= VMSTACK_DEPTH)
                return stack_fault(#{hereaddr});
        PUSH
    end
    def _c_pullstack(hereaddr)
        <<~PULL
            if (--vmstackptr < 0)
                return stack_fault(#{hereaddr});
            dyncallsite = #{hereaddr};
            entry = vmstack[vmstackptr];
            goto reentry;
        PULL
    end
end

class BaseInstruction

    include RASM_Directives
    include RASM_Instructions

    abstract_method :conditions? # Does this instruction support conditions?
    abstract_method :result? # Does this instruction write it's result?
    abstract_method :default_result? # Does this instruction write it's result by default?
    abstract_method :carry? # Can this instruction write the carry flag?
    abstract_method :zero? # Can this instruction write the zero flag?
    abstract_method :args # Describes the arguments/parameters this instruction takes (as an array of arrays of symbols)
    abstract_method :longs # How many longs does this instruction need in P8X32A machine code? (and thus in VM memory)
    abstract_method :pagemargin # How many longs this instruction must be away from the end of a page (for most instructions, this is 1)
    
    abstract_method :to_p8x32a # Output P8X32A machine code
    abstract_method :to_c # Output C code
    def caller? # Does this instruction make a call? (Marks the next instruction as dyncallable if neccessary (--switchcase))
        false
    end
    def pagemargin_if_always # Page margin if condition is :if_always
        pagemargin
    end
end

class ALUInstruction < BaseInstruction
    attr_reader :opcode
    attr_reader :c_impl

    def initialize(opcode,c_impl,default_wr)
        @opcode = opcode
        @c_impl = c_impl.gsub(/\n^\s*/,"\n    ")
        @default_wr = default_wr
    end

    def conditions?
        true
    end
    def result?
        true
    end
    def default_result?
        @default_wr
    end
    def carry?
        true
    end
    def zero?
        true
    end
    def args
        [[:cog_d,:cog_r],[:cog_d,:data]]
    end
    def longs
        1
    end
    def pagemargin
        1
    end
    
    def to_p8x32a(address,data)
        [_make_p8x32a_instruction(@opcode,
            data[:condition],data[:_wr],data[:_wc],data[:_wz],
            data[:pattern][1] == :data,
            _as_integer(data[:args][0],data[:_wr]),_as_integer(data[:args][1]))]
    end
    def to_c(address,data)
        _frame_c(address,
            c_impl,
            data[:condition],data[:_wr],data[:_wc],data[:_wz],
            _cogloc(_as_integer(data[:args][0],data[:_wr])),
            data[:pattern][1] == :data ?
            "#{_as_integer(data[:args][1]).to_s}" :
            _cogloc(_as_integer(data[:args][1])))
    end

end

ALUINSTRUCTIONS = {
    :abs     => [0b101010,%{dest = rasm_abs(source);
                            newnz = dest;
                            newc = source&(1<<31);},true],
    :absneg  => [0b101011,%{dest = 0-rasm_abs(source);
                            newnz = dest;
                            newc = source&(1<<31);},true], 
    :mov     => [0b101000,%{dest = source;
                            newnz = dest;
                            newc = source&(1<<31);},true],
    :neg     => [0b101001,%{dest = 0-source;
                            newnz = dest;
                            newc = source&(1<<31);},true],
    :negc    => [0b101100,%{dest = c ? 0-source : source;
                            newnz = dest;
                            newc = source&(1<<31);},true],
    :negnc   => [0b101101,%{dest = !c ? 0-source : source;
                            newnz = dest;
                            newc = source&(1<<31);},true],
    :negz    => [0b101110,%{dest = !nz ? 0-source : source;
                            newnz = dest;
                            newc = source&(1<<31);},true],
    :negnz   => [0b101111,%{dest = nz ? 0-source : source;
                            newnz = dest;
                            newc = source&(1<<31);},true],
    :muxc    => [0b011100,%{dest &= ~source;
                            if (c) dest |= source;
                            newnz = dest;
                            newc = fastparity32(dest);},true],
    :muxnc   => [0b011101,%{dest &= ~source;
                            if (!c) dest |= source;
                            newnz = dest;
                            newc = fastparity32(dest);},true],
    :muxz    => [0b011110,%{dest &= ~source;
                            if (!nz) dest |= source;
                            newnz = dest;
                            newc = fastparity32(dest);},true],
    :muxnz   => [0b011111,%{dest &= ~source;
                            if (nz) dest |= source;
                            newnz = dest;
                            newc = fastparity32(dest);},true],                
    :movs    => [0b010100,%{newc = dest < source;
                            dest &= ~(511);
                            dest |= source&(511);
                            newnz = dest;},true],
    :movd    => [0b010101,%{newc = dest < source;
                            dest &= ~(511<<9);
                            dest |= (source&(511))<<9;
                            newnz = dest;},true], 
    :movi    => [0b010110,%{newc = dest < source;
                            dest &= ~(511<<23);
                            dest |= (source&(511))<<23;
                            newnz = dest;},true],    
    :add     => [0b100000,%{newc = __builtin_add_overflow(dest,source,&dest);
                            newnz = dest;},true], 
    :addabs  => [0b100010,%{if (((vmlong_s)source)>=0) {
                                newc = __builtin_add_overflow(dest,source,&dest);
                            } else {
                                newc = dest<source;
                                dest -= source;
                            }
                            newnz = dest;},true],          
    :adds    => [0b110100,%{newc = __builtin_add_overflow((vmlong_s)dest,(vmlong_s)source,(vmlong_s*)&dest);
                            newnz = dest;},true],
    :addsx   => [0b110110,%{vmlong_s tmp;
                            newc  = __builtin_add_overflow((vmlong_s)(c?1:0),(vmlong_s)source,&tmp);
                            newc |= __builtin_add_overflow((vmlong_s)dest,tmp,(vmlong_s*)&dest);
                            newnz = nz || dest;},true],
    :addx    => [0b110010,%{vmlong tmp;
                            newc  = __builtin_add_overflow((vmlong)(c?1:0),source,&tmp);
                            newc |= __builtin_add_overflow(dest,tmp,&dest);
                            newnz = nz || dest;},true],
    :and    =>  [0b011000,%{newc = rasm_andparity(dest,source);
                            dest &= source;
                            newnz = dest;},true],
    :andp    => [:and,true],                
    :andn    => [0b011001,%{newc = rasm_andparity(dest,~source);
                            dest &= ~source;
                            newnz = dest;},true], 
    :test    => [:and,false],
    :testn   => [:andn,false],
    :or      => [0b011010,%{dest |= source;
                            newnz = dest;
                            newc = fastparity32(dest);},true],
    :ora     => [:or,true],                         
    :xor     => [0b011011,%{dest ^= source;
                            newnz = dest;
                            newc = fastparity32(dest);},true],                                            
    :sub     => [0b100001,%{newc = __builtin_sub_overflow(dest,source,&dest);
                            newnz = dest;},true],
    :subabs  => [0b100011,%{if (!(source&(1<<31))) {
                                newc = dest<source;
                                dest -= source;
                            } else {
                                vmlong tmp = rasm_abs(source);
                                newc = dest>=tmp;
                                dest -= tmp;
                            }
                            newnz = dest;},true],          
    :subs    => [0b110101,%{newc = __builtin_sub_overflow((vmlong_s)dest,(vmlong_s)source,(vmlong_s*)&dest);
                            newnz = dest;},true],
    :subsx   => [0b110111,%{vmlong_s tmp;
                            newc  = __builtin_sub_overflow((vmlong_s)dest,(vmlong_s)(c?1:0),&tmp);
                            newc |= __builtin_sub_overflow(tmp,(vmlong_s)source,(vmlong_s*)&dest);
                            newnz = nz || dest;},true],
    :subx    => [0b110011,%{vmlong tmp;
                            newc  = __builtin_sub_overflow(dest,(vmlong)(c?1:0),&tmp);
                            newc |= __builtin_sub_overflow(tmp,source,&dest);
                            newnz = nz || dest;},true],    
    :cmp     => [:sub,false],
    :cmps    => [0b110000,%{newc = (vmlong_s)dest<(vmlong_s)source;
                            dest = (vmlong)((vmlong_s)dest-(vmlong_s)source);
                            newnz = dest;},false],
    :cmpsub  => [0b111000,%{vmlong tmp = dest-source;
                            if (dest >= source) {
                                dest = tmp;
                                newc = true;
                            } else  newc = false;
                            newnz = tmp;},true],
    :cmpsx   => [0b110001,%{newc = ((vmlong_s)dest < ( ((vmlong_s)(c?1:0)) + (vmlong_s)source));
                            dest = (vmlong)((vmlong_s)dest - ( ((vmlong_s)(c?1:0)) + (vmlong_s)source));
                            newnz = nz || dest;},false],
    :cmpx    => [:subx,false],
    :shl     => [0b001011,%{newc = dest & (1<<31);
                            dest <<= (source&31);
                            newnz = dest;},true],
    :shr     => [0b001010,%{newc = dest & 1;
                            dest >>= (source&31);
                            newnz = dest;},true],
    :sar     => [0b001110,%{newc = dest & 1;
                            dest = (vmlong)((vmlong_s)dest >> (source&31));
                            newnz = dest;},true],
    :rol     => [0b001001,%{newc = dest & (1<<31);
                            dest = (dest << (source&31)) | (dest>>(-source&31));
                            newnz = dest;},true],
    :ror     => [0b001000,%{newc = dest & 1;
                            dest = (dest >> (source&31)) | (dest<<(-source&31));
                            newnz = dest;},true],
    :rev     => [0b001111,%{newc = dest & 1;
                            dest = rasm_rev(dest) >> (source&31);
                            newnz = dest;},true],
    :rcl     => [0b001101,%{newc = dest & (1<<31);
                            dest <<= (source&31);
                            if(c) dest |= (1<<(source&31))-1;
                            newnz = dest;},true],
    :rcr     => [0b001100,%{newc = dest & 1;
                            dest >>= (source&31);
                            if (c && (source&31)) dest |= 0-(1<<(-source&31));
                            newnz = dest;},true],
    :sumc    => [0b100100,%{vmlong_s s = (vmlong_s)source;
                            newc = __builtin_add_overflow((vmlong_s)dest,c?-s:s,(vmlong_s*)&dest);
                            newnz = dest;},true],
    :sumnc   => [0b100101,%{vmlong_s s = (vmlong_s)source;
                            newc = __builtin_add_overflow((vmlong_s)dest,c?s:-s,(vmlong_s*)&dest);
                            newnz = dest;},true],
    :sumz    => [0b100110,%{vmlong_s s = (vmlong_s)source;
                            newc = __builtin_add_overflow((vmlong_s)dest,nz?s:-s,(vmlong_s*)&dest);
                            newnz = dest;},true],    
    :sumnz   => [0b100111,%{vmlong_s s = (vmlong_s)source;
                            newc = __builtin_add_overflow((vmlong_s)dest,nz?-s:s,(vmlong_s*)&dest);
                            newnz = dest;},true],
    :max     => [0b010011,%{if (dest < source) {
                                newc = true;
                            } else {
                                newc = false;
                                dest = source;
                            }
                            newnz = source;},true],
    :maxs    => [0b010001,%{if ((vmlong_s)dest < (vmlong_s)source) {
                                newc = true;
                            } else {
                                newc = false;
                                dest = source;
                            }
                            newnz = source;},true],
    :min     => [0b010010,%{if (dest < source) {
                                newc = true;
                                dest = source;
                            } else {
                                newc = false;
                            }
                            newnz = source;},true],                                                   
    :mins    => [0b010000,%{if ((vmlong_s)dest < (vmlong_s)source) {
                                newc = true;
                                dest = source;
                            } else {
                                newc = false;
                            }
                            newnz = source;},true],
    
            

}
_register_many_instructions(ALUInstruction,ALUINSTRUCTIONS)

class FlagInstruction < BaseInstruction

    def initialize(setc,setz)
        @setc = setc
        @setz = setz
        @c_impl = 
            case @setc
            when true
                "c = true;"
            when false
                "c = false;"
            else ""
            end +
            case @setz
            when true
                "nz = false;"
            when false
                "nz = true;"
            else ""
            end
    end

    def result?
        false
    end
    def default_result?
        false
    end

    def conditions?
        true
    end

    def args
        [[]]
    end

    def carry?
        false
    end
    def zero?
        false
    end

    def longs
        1
    end
    def pagemargin
        1
    end

    def to_p8x32a(address,data)
        [_make_p8x32a_instruction(!!@setc == !!@setz ? 
            0b111000 : # CMPSUB
            0b100001 , # CMP
            data[:condition],false,!@setc.nil?,!@setz.nil?,
            true,
            _as_integer($labels[:vm_zero]),@setz ? 0 : 1)]
    end
    def to_c(address,data)
        _frame_c_min(address,data[:condition],@c_impl)
    end

end

FLAGINSTRUCTIONS = {
    :setc    => [true,nil],
    :setnc   => [false,nil],
    :clrc    => [:setnc],
    :setz    => [nil,true],
    :setnz   => [nil,false],
    :clrz    => [:setnz],
    :setncnz => [false,false],
    :setnznc => [:setncnz],
    :clrcz   => [:setncnz],
    :clrzc   => [:setncnz],
    :setncz  => [false,true],
    :setznc  => [:setncz],
    :setcnz  => [true,false],
    :setnzc  => [:setcnz],
    :setcz   => [true,true],
    :setzc   => [true,true],
}

_register_many_instructions(FlagInstruction,FLAGINSTRUCTIONS)

class HubInstruction < BaseInstruction

    attr_reader :opcode,:width,:zp,:write

    def initialize(opcode,width,zp,write)
        @opcode = opcode
        @width = width
        @zp = zp
        @write = write
    end

    def conditions?
        true
    end
    def result?
        false
    end
    def default_result?
        not @write
    end
    def carry?
        false
    end
    def zero?
        true
    end
    def args
        [if zp
            if write  
                [:cog_d,:hub_w]
            else
                [:cog_d,:hub_r]
            end 
        else
            [:cog_d,:cog_r]
        end]
    end
    def longs
        1
    end
    def pagemargin
        1
    end
    
    def to_p8x32a(address,data)
        [_make_p8x32a_instruction(@opcode,
            data[:condition],data[:_wr],data[:_wc],data[:_wz],
            @zp,
            _as_integer(data[:args][0],data[:_wr]),_as_integer(data[:args][1]))]
    end
    def to_c(address,data)
        _frame_c(address,
            if @write
                "wrhub#{@width}(dest,source,#{address});\n    newnz=(vm#{width})dest;"
            else
                "dest = rdhub#{@width}(source,#{address});\n    newnz=dest;"
            end,
            data[:condition],data[:_wr],data[:_wc],data[:_wz],
            _cogloc(_as_integer(data[:args][0],data[:_wr])),
            @zp ?
            "#{_as_integer(data[:args][1]).to_s}" :
            _cogloc(_as_integer(data[:args][1]))
        )
    end

end

[false,true].each do |write|
    [false,true].each do |zp|
        [[0b000010,"long"],[0b000001,"word"],[0b000000,"byte"]].each do |opcode,width|
            _register_instruction("#{write ? "wr" : "rd"}#{zp ? "z" : "h"}#{width}".to_sym,HubInstruction.new(opcode,width,zp,write))
        end
    end
end

class CodeReadInstruction < HubInstruction
    def initialize(opcode,width)
        super(opcode,width,false,false)
    end

    def to_c(address,data)
        _frame_c(address,
            <<~CODE ,
                if ((!(source & 0x80000000))||(source&0x7FFFFE00))
                    return coderead_fault(#{address});
                if(!rdcode#{@width}((source&0x1FF)|(#{address}&~0x1FF),dest))
                    return transfer_fault(#{address});
                newnz=dest;
            CODE
            data[:condition],data[:_wr],data[:_wc],data[:_wz],
            _cogloc(_as_integer(data[:args][0],data[:_wr])),
            _cogloc(_as_integer(data[:args][1])))
    end
end
[[0b000010,"long"],[0b000001,"word"],[0b000000,"byte"]].each do |opcode,width|
    _register_instruction("rdcode#{width}".to_sym,CodeReadInstruction.new(opcode,width))
end

class KernalInstruction < BaseInstruction
    attr_reader :label
    attr_reader :c_impl
    attr_reader :return_lbl
    attr_reader :is_jumptable

    def initialize(c_impl,label,return_lbl=:vm_return,is_jumptable=false)
        @label = label
        @c_impl = c_impl
        @return_lbl = return_lbl
        @is_jumptable = is_jumptable
    end

    def conditions?
        true
    end
    def result?
        false
    end
    def default_result?
        false
    end
    def carry?
        false
    end
    def zero?
        false
    end
    def args
        [[]]
    end
    def longs
        1
    end
    def pagemargin
        1
    end
    
    def to_p8x32a(address,data,force_native_condition=false)
        if force_native_condition or longs == 1 or data[:condition] == CONDITIONS[:if_always]
            [_make_p8x32a_instruction(0b010111,
                data[:condition],!!return_lbl,false,false,
                !is_jumptable,
                $labels[(return_lbl or :vm_return)].loc!,$labels[label].loc!)]
        elsif not return_lbl # Emulated condition
            [_make_p8x32a_instruction(0b010111,
                CONDITIONS[:if_always],false,false,false,
                !is_jumptable,
                data[:condition].p8x32a,$labels["#{label}_cond".to_sym].loc!)]
        else 
            raise
        end
    end

    def to_c(address,data)
        ccode = @c_impl
        ccode = ccode.call(address,data) if ccode.is_a? Proc and ccode.lambda?
        _frame_c_min(address,data[:condition],ccode)
    end

end

class SJMPInstruction < ALUInstruction

    def initialize
        super(ALUINSTRUCTIONS[:movs][0],"",true)
    end

    def result?
        false
    end
    def carry?
        false
    end
    def zero?
        false
    end
    def args
        [[:vm_r],[:cog_r]]
    end

    def to_p8x32a(address,data)
        if data[:pattern][0] == :vm_r and address & ~0x1ff != _as_integer(data[:args][0]) & ~0x1ff
            raise RASM_StateError, "SJMP out of range"
        end
        [_make_p8x32a_instruction(@opcode,
            data[:condition],data[:_wr],data[:_wc],data[:_wz],
            data[:pattern][0] == :vm_r,
            _as_integer($labels[:vm_lmmpc]),511 & _as_integer(data[:args][0]))]
    end

    def to_c(address,data)
        if data[:pattern][0] == :vm_r
            _frame_c_min(address,data[:condition],"goto #{_c_label(_as_integer(data[:args][0]))};")
        else
            _frame_c_min(address,data[:condition],_c_dynjmp(address,"((#{address}&(~0x1ff))|(#{_cogloc(_as_integer(data[:args][0]))}&0x1ff))"))
        end
    end

    def pagemargin_if_always
        0
    end
end
_register_instruction(:sjmp,SJMPInstruction.new)
_register_instruction(:stail,SJMPInstruction.new)

class NextpageInstruction < KernalInstruction

    def initialize(label=:vm_nextpage)
        super(lambda do |a,d|
            addr = (a+511)&~511
            "goto #{_c_label(addr)};"
        end,label,false)
    end

    def to_p8x32a(address,data)
        if address & 508 == 508 and data[:condition] != CONDITIONS[:if_always]
            raise RASM_StateError,"Conditional nextpage at end of page"
        else
            super
        end
    end

    def pagemargin
        0
    end

end

_register_instruction(:nextpage,NextpageInstruction.new)
_register_instruction(:nextfast,NextpageInstruction.new(:vm_nextfast))







end